/* eckcdsa/eckcdsa_iso.c */
/* 
 * EC-KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include <openssl/evp.h>
#include "eckcdsa_lcl.h"

static ECKCDSA_METHOD iso_eckcdsa_meth = {
	"ECKCDSA ISO method",
	eckcdsa_do_sign,
	eckcdsa_sign_setup,
	eckcdsa_do_verify,
	eckcdsa_gen_param,
	eckcdsa_gen_key,
};

/* ISO method */
const ECKCDSA_METHOD * ECKCDSA_iso_method()
{
	return &iso_eckcdsa_meth;
}

/* ISO method functions */
int eckcdsa_gen_param(ECKCDSA *eckcdsa, const EVP_MD * evp_md, int nid)
{
	eckcdsa->evp_md = evp_md;
	if (eckcdsa->evp_md == NULL) eckcdsa->evp_md = EVP_sha1();
	if (eckcdsa->evp_md == NULL) return 0;

	/* clean up old informations */
	if (eckcdsa->group != NULL) EC_GROUP_free(eckcdsa->group);
	if (eckcdsa->priv_key != NULL) BN_clear_free(eckcdsa->priv_key);
	if (eckcdsa->pub_key != NULL) EC_POINT_free(eckcdsa->pub_key);
	if (eckcdsa->k != NULL) BN_clear_free(eckcdsa->k);
	if (eckcdsa->r != NULL) {
		OPENSSL_cleanse(eckcdsa->r, eckcdsa->rlen);
		OPENSSL_free(eckcdsa->r);
    }
	eckcdsa->group = NULL;
	eckcdsa->priv_key = NULL;
	eckcdsa->pub_key = NULL;
	eckcdsa->k = NULL;
	eckcdsa->r = NULL;
	eckcdsa->rlen = 0;

	/* make new one */
	eckcdsa->group = EC_GROUP_new_by_curve_name(nid);
	if (eckcdsa->group == NULL) return 0;

	return 1;
}

/* EC_KEY_generate_key()는 ECDSA 방식이다. 그러므로 여기서 key generation은
 * 따로 만들어야 한다. */
int eckcdsa_gen_key(ECKCDSA *eckcdsa)
{
	int ok = 0;
	int tmp_sensitive = 0;
	BIGNUM *priv_key = NULL;
	EC_POINT *pub_key = NULL;
	EC_GROUP *group = eckcdsa->group;
	BIGNUM *order = NULL;
	BIGNUM *tmp = NULL;
	BN_CTX *ctx = NULL;

	/* some mandatory check */
	if (!group) {
		fprintf(stderr, "eckcdsa_gen_key(): group not defined.\n");
		return 0;
	}

	/* clean up and alloc */
	priv_key = eckcdsa->priv_key;
	pub_key = eckcdsa->pub_key;
	eckcdsa->priv_key = NULL;
	eckcdsa->pub_key = NULL;

	if (priv_key == NULL) {
		priv_key = BN_new();
		if (priv_key == NULL) goto keygen_err;
	} else {
		BN_clear(priv_key);
	}
	if (eckcdsa->pub_key == NULL) {
		pub_key = EC_POINT_new(group);
		if (pub_key == NULL) goto keygen_err;
	}
	if ((ctx = BN_CTX_new()) == NULL) goto keygen_err;

	BN_CTX_start(ctx);
	order = BN_CTX_get(ctx);
	tmp = BN_CTX_get(ctx);
	if (tmp == NULL) goto keygen_err;

	/* NOTE: We can reuse existing values of K and R even if the key pair
	 * changed. */

	if (!EC_GROUP_get_order(group, order, ctx)) goto keygen_err;

	/* 14888-3:2006 step 6.5.2 */
	do
		if (!BN_rand_range(priv_key, order)) goto keygen_err;
	while (BN_is_zero(priv_key));

	if (!BN_mod_inverse(tmp, priv_key, order, ctx)) goto keygen_err;
	tmp_sensitive = 1;
	if (!EC_POINT_mul(group, pub_key, tmp, NULL, NULL, ctx)) goto keygen_err;
	BN_clear(tmp); tmp_sensitive = 0;

	eckcdsa->priv_key = priv_key;
	eckcdsa->pub_key = pub_key;
	priv_key = NULL;
	pub_key = NULL;

	ok = 1;

keygen_err:
	if (priv_key != NULL) BN_clear_free(priv_key);
	if (pub_key != NULL) EC_POINT_free(pub_key);
	if (ctx != NULL) {
		if (tmp_sensitive) BN_clear(tmp);
		BN_CTX_end(ctx);
		BN_CTX_free(ctx);
	}

	return ok;
}

