/* eckcdsa/eckcdsa_iso_sign.c */
/* 
 * EC-KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include <string.h>
#include <openssl/evp.h>
#include "eckcdsa_lcl.h"

ECKCDSA_SIG * eckcdsa_do_sign(const unsigned char *mesg, int mlen,
		ECKCDSA *eckcdsa)
{
	ECKCDSA_SIG *sig = NULL;
	BN_CTX *ctx = NULL;
	EVP_MD_CTX *md_ctx = NULL;
	const EVP_MD *evp_md = eckcdsa->evp_md;
	unsigned char *R = NULL;
	BIGNUM *K = NULL, *S = NULL;
	size_t rlen = 0, tmplen;
	unsigned char buf[1024]; // TODO: 경제적으로 */
	BIGNUM *V=NULL,*tmp = NULL;
	BIGNUM *order = NULL;
	size_t mdlen;
	int i;

	/* sanity check */
	if (!evp_md) goto err;
	if (!eckcdsa->group || !eckcdsa->priv_key || !eckcdsa->pub_key) {
		goto err;
	}

	/* alloc */
	if ((S = BN_new()) == NULL) goto err;
	if ((ctx = BN_CTX_new()) == NULL) goto err;
	BN_CTX_start(ctx);
	order = BN_CTX_get(ctx);
	V = BN_CTX_get(ctx);
	tmp = BN_CTX_get(ctx);
	if (tmp == NULL) goto err;
	if (!EC_GROUP_get_order(eckcdsa->group, order, ctx)) goto err;

	/* check precomputed signature parts */
	if (!eckcdsa->k || !eckcdsa->rlen) {
		if (!ECKCDSA_sign_setup(eckcdsa, ctx, &K, &R, &rlen)) goto err;
	} else {
		K = eckcdsa->k;
		R = eckcdsa->r;
		rlen = eckcdsa->rlen;
		eckcdsa->k = NULL;
		eckcdsa->r = NULL;
		eckcdsa->rlen = 0;
	}

	// 14888-3 step 6.5.3.5: computing the assignment
	/* Y' = leftmost l bits of FE2BS(Yx)||FE2BS(Yy) */
	tmplen = EC_POINT_point2oct(eckcdsa->group, eckcdsa->pub_key,
			POINT_CONVERSION_UNCOMPRESSED, buf, 1024, ctx);

	/* H = h(Y'||mesg) */
	md_ctx = EVP_MD_CTX_create();
	if (md_ctx == NULL) goto err;
	if (!EVP_DigestInit_ex(md_ctx, evp_md, NULL)) goto err;
	/* update with Y' */
	/* EC_POINT_point2oct() palces conversion type in the first bytes.
	 * Discard it. */
	if (!EVP_DigestUpdate(md_ctx, buf+1, tmplen-1)) goto err;
	/* update with mesg */
	if (!EVP_DigestUpdate(md_ctx, mesg, mlen)) goto err;
	/* md final */
	if (!EVP_DigestFinal_ex(md_ctx, buf, &mdlen)) goto err;

	/* V = BS2I(R xor H) mod q */
	// TODO: assertion that rlen = mdlen */
	for (i = mdlen-1; i >= 0; i--) {
		buf[i] ^= R[i];
	}
	if (!BN_bin2bn(buf, mdlen, tmp)) goto err;
	if (!BN_mod(V, tmp, order, ctx)) goto err;

	// 14888-3 step 6.5.3.6: computing the second part of the signature
	/* S = X(K - V) mod q */
	while (BN_cmp(K,V) < 0)
		BN_add(K,K,order);
	BN_sub(S,K,V);
	if (!BN_mod_mul(S, S, eckcdsa->priv_key, order, ctx)) goto err;

	// 14888-3 step 6.5.3.7: constructing the appendix
	if ((sig = ECKCDSA_SIG_new()) == NULL) goto err;
	sig->r = R;
	sig->rlen = rlen;
	sig->s = S;
	R = NULL;
	rlen = 0;
	S = NULL;

err:
	if (ctx) {
		BN_CTX_end(ctx);
		BN_CTX_free(ctx);
	}
	if (K) BN_clear_free(K);
	if (R) {
		OPENSSL_cleanse(R, rlen);
		OPENSSL_free(R);
	}
	if (S) BN_clear_free(S);
	if (md_ctx != NULL) EVP_MD_CTX_destroy(md_ctx);

	return sig;
}

int eckcdsa_sign_setup(ECKCDSA *eckcdsa, BN_CTX *ctx_in, BIGNUM **kp,
		unsigned char **rp, size_t *rlenp)
{
	BIGNUM *K = NULL, *order = NULL;
	unsigned char *R = NULL;
	BN_CTX *ctx = NULL;
	EC_GROUP *group = NULL;
	EC_POINT *pi = NULL;
	const EVP_MD *evp_md = eckcdsa->evp_md;
	unsigned char buf[1024]; // TODO: 경제적으로
	size_t len;
	int ret = 0;

	ctx = ctx_in;
	BN_CTX_start(ctx);

	/* sanity check again for safety */
	if (!evp_md) goto err;
	if (!eckcdsa->group || !eckcdsa->priv_key || !eckcdsa->pub_key) {
		goto err;
	}

	/* alloc */
	if ((K = BN_new()) == NULL) goto err;
	if (ctx_in == NULL) goto err;
	order = BN_CTX_get(ctx);
	if (order == NULL) goto err;

	group = eckcdsa->group;
	if ((pi = EC_POINT_new(group)) == NULL) goto err;
	if (!EC_GROUP_get_order(group, order, ctx)) goto err;

	// 14888-3 step 6.5.3.1: producing the randomizer
	do
		if (!BN_rand_range(K, order)) goto err;
	while (BN_is_zero(K));

	// 14888-3 step 6.5.3.2: producing the pre-signature
	if (!EC_POINT_mul(group, pi, K, NULL, NULL, ctx)) goto err;

	// 14888-3 step 6.5.3.3: preparing the message for signing

	// 14888-3 step 6.5.3.4: computing the witness
	len = EC_POINT_point2oct(group, pi,
			POINT_CONVERSION_COMPRESSED, buf, 1024, ctx);
	/* EC_POINT_point2oct() palces conversion type in the first byte and
	 * compressed y coord in the last byte.  Discard them. */
	if (!EVP_Digest(buf, len-2, buf+1, rlenp, evp_md, NULL)) goto err;
	R = (unsigned char *)OPENSSL_malloc(*rlenp);
	if (!memcpy(R, buf, *rlenp)) goto err;

	// collecting
	*kp = K;
	K = NULL;
	*rp = R;
	R = NULL;

	ret = 1;
err:
	BN_CTX_end(ctx);
	if (K) BN_clear_free(K);
	if (pi) EC_POINT_free(pi);
	if (R) OPENSSL_free(R);

	return ret;
}

