/* eckcdsa/eckcdsa_iso_vrf.c */
/* 
 * EC-KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include <openssl/evp.h>
#include "eckcdsa_lcl.h"

int eckcdsa_do_verify(const unsigned char *mesg, int mlen,
		ECKCDSA_SIG *sig, ECKCDSA *eckcdsa)
{
	const EVP_MD *evp_md = eckcdsa->evp_md;
	BN_CTX *ctx = NULL;
	EVP_MD_CTX *md_ctx = NULL;
	unsigned char *R = NULL;
	size_t rlen, mdlen, tmplen;
	unsigned char buf[1024]; // TODO: 경제적으로 */
	BIGNUM *S = NULL, *V = NULL;
	BIGNUM *order = NULL;
	BIGNUM *tmp = NULL;
	EC_POINT *pi = NULL;
	int i;
	int ret = -1;

	/* sanity check */
	if (!evp_md) goto err;
	if (!eckcdsa->group || !eckcdsa->pub_key) {
		goto err;
	}

	/* alloc */
	if ((ctx = BN_CTX_new()) == NULL) goto err;
	BN_CTX_start(ctx);
	order = BN_CTX_get(ctx);
	V = BN_CTX_get(ctx);
	tmp = BN_CTX_get(ctx);
	if (tmp == NULL) goto err;
	if (!EC_GROUP_get_order(eckcdsa->group, order, ctx)) goto err;
	if ((pi = EC_POINT_new(eckcdsa->group)) == NULL) goto err;

	// 14888-3 step 6.5.4.1: retrieving the witness
	R = sig->r;
	rlen = sig->rlen;
	S = sig->s;
	mdlen = EVP_MD_size(evp_md);
	/* size check */
	if (!R || !rlen || rlen != mdlen) goto err;
	/* range check */
	if (!S || BN_is_zero(S) || BN_cmp(S, order) >= 0) goto err;

	// 14888-3 step 6.5.4.2: preparing the message for verification

	// 14888-3 step 6.5.4.3: retrieving the assignment (6.5.3.5)
	/* Y' = leftmost l bits of FE2BS(Yx)||FE2BS(Yy) */
	tmplen = EC_POINT_point2oct(eckcdsa->group, eckcdsa->pub_key,
			POINT_CONVERSION_UNCOMPRESSED, buf, 1024, ctx);

	/* H = h(Y'||mesg) */
	md_ctx = EVP_MD_CTX_create();
	if (md_ctx == NULL) goto err;
	if (!EVP_DigestInit_ex(md_ctx, evp_md, NULL)) goto err;
	/* update with Y' */
	/* EC_POINT_point2oct() palces conversion type in the first bytes.
	 * Discard it. */
	if (!EVP_DigestUpdate(md_ctx, buf+1, tmplen-1)) goto err;
	/* update with mesg */
	if (!EVP_DigestUpdate(md_ctx, mesg, mlen)) goto err;
	/* md final */
	if (!EVP_DigestFinal_ex(md_ctx, buf, &mdlen)) goto err;

	/* V = BS2I(R xor H) mod q */
	for (i = mdlen-1; i >= 0; i--) {
		buf[i] ^= R[i];
	}
	if (!BN_bin2bn(buf, mdlen, tmp)) goto err;
	if (!BN_mod(V, tmp, order, ctx)) goto err;

	// 14888-3 step 6.5.4.4: recomputing the pre-signature
	if (!EC_POINT_mul(eckcdsa->group, pi, V, eckcdsa->pub_key, S, ctx))
		goto err;

	// 14888-3 step 6.5.4.5: recomputing the witness
	tmplen = EC_POINT_point2oct(eckcdsa->group, pi,
			POINT_CONVERSION_COMPRESSED, buf, 1024, ctx);
	/* EC_POINT_point2oct() palces conversion type in the first byte and
	 * compressed y coord in the last byte.  Discard them. */
	if (!EVP_Digest(buf, tmplen-2, buf+1, NULL, evp_md, NULL)) goto err;

	// 14888-3 step 6.5.4.6: verifying the witness
	if (memcmp(buf, R, rlen)) ret = 0; goto err;

	ret = 1;
err:
	if (ctx) {
		BN_CTX_end(ctx);
		BN_CTX_free(ctx);
	}
	if (md_ctx != NULL) EVP_MD_CTX_destroy(md_ctx);
	if (pi) EC_POINT_free(pi);

	return ret;
}

