/* eckcdsa/eckcdsa_lib.c */
/* 
 * EC-KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include "eckcdsa_lcl.h"

/* originally in crypto/asn1/t_pkey.c */
static int print(BIO *fp,const char *str, const BIGNUM *num,
		unsigned char *buf,int off);

static const ECKCDSA_METHOD * default_ECKCDSA_method = NULL;

const ECKCDSA_METHOD * ECKCDSA_get_default_method()
{
	if (default_ECKCDSA_method == NULL)
		default_ECKCDSA_method = ECKCDSA_iso_method();
    return default_ECKCDSA_method;
}

ECKCDSA_SIG * ECKCDSA_do_sign(const unsigned char *mesg, size_t mlen,
		ECKCDSA *eckcdsa)
{
	return eckcdsa->meth->eckcdsa_do_sign(mesg, mlen, eckcdsa);
}

int ECKCDSA_sign_setup(ECKCDSA *eckcdsa, BN_CTX *ctx_in, BIGNUM **kp,
		unsigned char **rp, size_t *rlenp)
{
	return eckcdsa->meth->eckcdsa_sign_setup(eckcdsa, ctx_in, kp, rp, rlenp);
}

int ECKCDSA_do_verify(const unsigned char *mesg, size_t mlen,
		ECKCDSA_SIG *sig, ECKCDSA *eckcdsa)
{
	return eckcdsa->meth->eckcdsa_do_verify(mesg, mlen, sig, eckcdsa);
}

int ECKCDSA_generate_parameters(ECKCDSA *eckcdsa,
		const EVP_MD * evp_md, int nid)
{
	return eckcdsa->meth->eckcdsa_gen_param(eckcdsa, evp_md, nid);
}

int ECKCDSA_generate_key(ECKCDSA *eckcdsa)
{
	return eckcdsa->meth->eckcdsa_gen_key(eckcdsa);
}

ECKCDSA *ECKCDSA_new(void)
{
    ECKCDSA *ret = NULL;

    ret = (ECKCDSA*) OPENSSL_malloc(sizeof(ECKCDSA));
    if (ret == NULL) {
		return NULL;
    }

	ret->evp_md = NULL;
	ret->group = NULL;
	ret->priv_key = NULL;
	ret->pub_key = NULL;
	ret->conv_form = POINT_CONVERSION_UNCOMPRESSED;
	ret->k = NULL;
	ret->r = NULL;
	ret->rlen = 0;
	ret->meth = ECKCDSA_get_default_method();

	return ret;
}

void ECKCDSA_free(ECKCDSA *to)
{
	if (to->group != NULL) EC_GROUP_free(to->group);
	if (to->priv_key != NULL) BN_clear_free(to->priv_key);
	if (to->pub_key != NULL) EC_POINT_free(to->pub_key);
	if (to->k != NULL) BN_clear_free(to->k);
	if (to->r != NULL) {
		OPENSSL_cleanse(to->r, to->rlen);
		OPENSSL_free(to->r);
    }
    OPENSSL_free(to);
}

ECKCDSA_SIG * ECKCDSA_SIG_new(void)
{
    ECKCDSA_SIG *ret = NULL;

    ret = (ECKCDSA_SIG*) OPENSSL_malloc(sizeof(ECKCDSA_SIG));
	if (ret == NULL) {
		return NULL;
	}

	ret->r = NULL;
	ret->rlen = 0;
	ret->s = NULL;

	return ret;
}

void ECKCDSA_SIG_free(ECKCDSA_SIG *to)
{
	if (to->r != NULL) {
		OPENSSL_cleanse(to->r, to->rlen);
		OPENSSL_free(to->r);
	}
	if (to->s != NULL) BN_clear_free(to->s);
	OPENSSL_free(to);
}

/* copied EC_KEY_print() and modified */
void ECKCDSA_print(BIO *bp, const ECKCDSA *x, int off)
	{
	unsigned char *buffer=NULL;
	size_t	buf_len=0, i;
	int     ret=0;
	BIGNUM  *pub_key=NULL, *order=NULL;
	BN_CTX  *ctx=NULL;
	const EC_GROUP *group;
	const EC_POINT *public_key;
	const BIGNUM *priv_key;
 
	if (x == NULL || (group = x->group) == NULL)
		{
		goto err;
		}

	if ((order = BN_new()) == NULL)
		{
		goto err;
		}
	if (!EC_GROUP_get_order(group, order, NULL))
		{
		goto err;
		}
 
	if (x->pub_key) {
		public_key = x->pub_key;
		if ((pub_key = EC_POINT_point2bn(group, public_key,
						x->conv_form, NULL, ctx)) == NULL)
			goto err;
		buf_len = (size_t)BN_num_bytes(pub_key);
	}

	priv_key = x->priv_key;
	if (priv_key != NULL)
		{
		if ((i = (size_t)BN_num_bytes(priv_key)) > buf_len)
			buf_len = i;
		}

	buf_len += 10;
	if ((buffer = OPENSSL_malloc(buf_len)) == NULL)
		{
		goto err;
		}

	if (priv_key != NULL)
		{
		if (!BIO_indent(bp, off, 128))
			goto err;
		if (BIO_printf(bp, "Private-Key: (%d bit)\n", 
			BN_num_bits(order)) <= 0) goto err;
		}
  
	if ((priv_key != NULL) && !print(bp, "priv:", priv_key, 
		buffer, off))
		goto err;
	if ((pub_key != NULL) && !print(bp, "pub: ", pub_key,
		buffer, off))
		goto err;
	if (!ECPKParameters_print(bp, group, off))
		goto err;
	ret=1;
err:
	if (pub_key) 
		BN_free(pub_key);
	if (order)
		BN_free(order);
	if (ctx)
		BN_CTX_free(ctx);
	if (buffer != NULL)
		OPENSSL_free(buffer);
	}

/* following functions originally in crypto/asn1/t_pkey.c */
static int print(BIO *bp, const char *number, const BIGNUM *num, unsigned char *buf,
	     int off)
	{
	int n,i;
	const char *neg;

	if (num == NULL) return(1);
	neg = (BN_is_negative(num))?"-":"";
	if(!BIO_indent(bp,off,128))
		return 0;
	if (BN_is_zero(num))
		{
		if (BIO_printf(bp, "%s 0\n", number) <= 0)
			return 0;
		return 1;
		}

	if (BN_num_bytes(num) <= BN_BYTES)
		{
		if (BIO_printf(bp,"%s %s%lu (%s0x%lx)\n",number,neg,
			(unsigned long)num->d[0],neg,(unsigned long)num->d[0])
			<= 0) return(0);
		}
	else
		{
		buf[0]=0;
		if (BIO_printf(bp,"%s%s",number,
			(neg[0] == '-')?" (Negative)":"") <= 0)
			return(0);
		n=BN_bn2bin(num,&buf[1]);
	
		if (buf[1] & 0x80)
			n++;
		else	buf++;

		for (i=0; i<n; i++)
			{
			if ((i%15) == 0)
				{
				if(BIO_puts(bp,"\n") <= 0
				   || !BIO_indent(bp,off+4,128))
				    return 0;
				}
			if (BIO_printf(bp,"%02x%s",buf[i],((i+1) == n)?"":":")
				<= 0) return(0);
			}
		if (BIO_write(bp,"\n",1) <= 0) return(0);
		}
	return(1);
	}
