/* eckcdsa/eckcdsatest.c */
/* 
 * EC-KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

/* Until the key-gen callbacks are modified to use newer prototypes, we allow
 * deprecated functions for openssl-internal code */
#ifdef OPENSSL_NO_DEPRECATED
#undef OPENSSL_NO_DEPRECATED
#endif

#include <stdio.h>

#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include "eckcdsa.h"

static const unsigned char mesg[]="message to be signed";

static const char rnd_seed[] = "string to make the random number generator think it has entropy";

int main()
{
	int ret = 0;
	ECKCDSA *eckcdsa=NULL;
	ECKCDSA_SIG *sig = NULL;
	int nid;
	BIO *out;

	out = BIO_new_fp(stdout, BIO_NOCLOSE);

	CRYPTO_malloc_debug_init();
	CRYPTO_dbg_set_options(V_CRYPTO_MDEBUG_ALL);
	CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);

	ERR_load_crypto_strings();
	RAND_seed(rnd_seed, sizeof rnd_seed);

	/* openssl의 string 입출력은 BIO module을 통해서 이루어짐 */
	BIO_printf(out,"testing EC-KCDSA\n");

	// alloc ECKCDSA
	if ((eckcdsa = ECKCDSA_new()) == NULL) goto err;

	// generate parameters
	/* DSA와 KCDSA의 경우와는 다르게 타원곡선을 사용하는 ECDSA와 EC-KCDSA의
	 * 경우 domain parameter를 생성하는 것은 훨씬 복잡한 과정을 거친다. 따라서
	 * 대부분의 경우 미리 생성된 curve parameter들을 골라서 사용하게 된다. */
	nid = NID_X9_62_prime192v1;
	BIO_printf(out, "curve %s: \n", OBJ_nid2sn(nid));
	ECKCDSA_generate_parameters(eckcdsa, EVP_sha1(), nid);

	// check parameters

	// generate key pair
	ECKCDSA_generate_key(eckcdsa);
	ECKCDSA_print(out, eckcdsa, 0);

	// sign
	sig = ECKCDSA_do_sign(mesg, 20, eckcdsa);
	if (sig == NULL) {
		fprintf(stderr, "signing failed\n");
		goto err;
	}

	// verify
	if (!ECKCDSA_do_verify(mesg, 20, sig, eckcdsa)) {
		fprintf(stderr, "verification failed\n");
		goto err;
	}

	ret = 1;
	BIO_printf(out, "test done\n");

err:
	/* 원래 ERR_print_errors()에 의해서 모든 에러 메시지가 출력되어야 하지만,
	 * kcdsa 관련한 에러 메시지는 포함되어 있지 않다. 따라서 ret == 0일 때, 즉
	 * 서명 검증이 실패한 경우 따로 에러 메시지를 출력한다. */
	if (!ret)
		ERR_print_errors(out);

	/* application specific garbage collecting */
	if (!ret)
		fprintf(stderr, "error\n");
	if (sig != NULL) ECKCDSA_SIG_free(sig);
	if (eckcdsa != NULL) ECKCDSA_free(eckcdsa);

	CRYPTO_cleanup_all_ex_data();
	ERR_remove_state(0);
	ERR_free_strings();
	CRYPTO_mem_leaks(out);
	if (out != NULL)
		{
		BIO_free(out);
		out = NULL;
		}
	exit(!ret);
	return(0);
}
