/* kcdsa/kcdsa.h */
/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#ifndef KCDSA_H
#define KCDSA_H

#undef OPENSSL_NO_KCDSA

#ifndef OPENSSL_NO_BIO
#include <openssl/bio.h>
#endif

#include <openssl/bn.h>

/* ISO 14888-3에서는 seed의 최대 크기에 대해서는 언급이 없고 최소 크기를 q의
 * bit size만큼이라고만 규정하고 있음. q의 크기는 160+32*j (0 <= j <= 3)라고
 * 정의돼 있으므로 최대 256 bits (32 bytes)이다. */
#define KCDSA_MAX_Q_LENGTH	32

#define KCDSA_MAX_MODULUS_BITS	2048

#define KCDSA_FLAG_CACHE_MONT_P		0x01
#define KCDSA_FLAG_NO_EXP_CONSTTIME	0x02

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct kcdsa_st KCDSA;
typedef struct kcdsa_method KCDSA_METHOD;

typedef struct KCDSA_SIG_st
	{
	unsigned char *r;
	size_t rlen; /* r의 크기 (byte 단위) */
	BIGNUM *s;
	} KCDSA_SIG;


KCDSA_SIG * KCDSA_SIG_new(void);
void	KCDSA_SIG_free(KCDSA_SIG *a);

/* KCDSA_sign()과 KCDSA_vefiry()는 ASN.1 encoding, decoding을 다루는
 * 함수들이다. 따라서 ASN.1 coding과 상관이 없는 함수는 _do_가 첨가된 이름들을
 * 갖는 것이 기존의 OpenSSL에 어울리기 쉽다. */
KCDSA_SIG * KCDSA_do_sign(const unsigned char *mesg,int mlen,KCDSA *kcdsa);
/* the following returns -1 on error */
int	KCDSA_do_verify(const unsigned char *mesg,int mlen, KCDSA_SIG *sig,
		KCDSA *kcdsa);

const KCDSA_METHOD *KCDSA_iso_method(void);
const KCDSA_METHOD *KCDSA_tta_method(void);

void	KCDSA_set_default_method(const KCDSA_METHOD *);
const KCDSA_METHOD *KCDSA_get_default_method(void);
int	KCDSA_set_method(KCDSA *kcdsa, const KCDSA_METHOD *);

KCDSA *	KCDSA_new(void);
KCDSA *	KCDSA_new_method(ENGINE *engine);
void	KCDSA_free (KCDSA *r);
/* the following returns -1 on error */
int	KCDSA_sign_setup( KCDSA *kcdsa,BN_CTX *ctx_in,BIGNUM **kp,unsigned char **rp, size_t *rlenp);

int	KCDSA_generate_parameters_ex(KCDSA *kcdsa, int pbits, int qbits,
		unsigned char *seed,int seed_len, const EVP_MD *evp_md,
		int *counter_ret, unsigned long *h_ret, BN_GENCB *cb);

int	KCDSA_generate_key(KCDSA *a);

#ifndef OPENSSL_NO_BIO
int	KCDSAparams_print(BIO *bp, const KCDSA *x);
int	KCDSA_print(BIO *bp, const KCDSA *x, int off);
#endif
#ifndef OPENSSL_NO_FP_API
int	KCDSAparams_print_fp(FILE *fp, const KCDSA *x);
int	KCDSA_print_fp(FILE *bp, const KCDSA *x, int off);
#endif

#define kcdsa_num_prime_checks 50
/* Primality test. 50 rounds of Rabin-Miller */
#define KCDSA_is_prime(n, callback, cb_arg) \
	BN_is_prime(n, kcdsa_num_prime_checks, callback, NULL, cb_arg)

/* BEGIN ERROR CODES */
/* The following lines are auto generated by the script mkerr.pl. Any changes
 * made after this point may be overwritten when the script is next run.
 */
void ERR_load_KCDSA_strings(void);

/* Error codes for the KCDSA functions. */

#include <openssl/err.h>

/* Function codes. */
#define KCDSA_F_KCDSA_NEW_METHOD			 100
#define KCDSA_F_KCDSA_SIGN_SETUP			101
#define KCDSA_F_KCDSA_DO_VERIFY				102

/* Reason codes. */
#define KCDSA_R_MISSING_PARAMETERS			 100

#ifdef  __cplusplus
}
#endif
#endif
