/* kcdsa/kcdsa_asn1.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 * slight modification from DSA source code from openssl-0.9.8k
 */

#include <stdio.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/bn.h>

#include "kcdsa_lcl.h"

/* Override the default new methods */
static int sig_cb(int operation, ASN1_VALUE **pval, const ASN1_ITEM *it)
{
	if(operation == ASN1_OP_NEW_PRE) {
		KCDSA_SIG *sig;
		sig = OPENSSL_malloc(sizeof(KCDSA_SIG));
		sig->r = NULL;
		sig->rlen = 0;
		sig->s = NULL;
		*pval = (ASN1_VALUE *)sig;
		if(sig) return 2;
		//KCDSAerr(KCDSA_F_SIG_CB, ERR_R_MALLOC_FAILURE);
		return 0;
	}
	return 1;
}

ASN1_SEQUENCE_cb(KCDSA_SIG, sig_cb) = {
	ASN1_SIMPLE(KCDSA_SIG, r, CBIGNUM),
	ASN1_SIMPLE(KCDSA_SIG, s, CBIGNUM)
} ASN1_SEQUENCE_END_cb(KCDSA_SIG, KCDSA_SIG)

IMPLEMENT_ASN1_ENCODE_FUNCTIONS_const_fname(KCDSA_SIG,KCDSA_SIG,KCDSA_SIG)

/* Override the default free and new methods */
static int kcdsa_cb(int operation, ASN1_VALUE **pval, const ASN1_ITEM *it)
{
	if(operation == ASN1_OP_NEW_PRE) {
		*pval = (ASN1_VALUE *)KCDSA_new();
		if(*pval) return 2;
		return 0;
	} else if(operation == ASN1_OP_FREE_PRE) {
		KCDSA_free((KCDSA *)*pval);
		*pval = NULL;
		return 2;
	}
	return 1;
}

ASN1_SEQUENCE_cb(KCDSAPrivateKey, kcdsa_cb) = {
	ASN1_SIMPLE(KCDSA, version, LONG),
	ASN1_SIMPLE(KCDSA, p, BIGNUM),
	ASN1_SIMPLE(KCDSA, q, BIGNUM),
	ASN1_SIMPLE(KCDSA, g, BIGNUM),
	ASN1_SIMPLE(KCDSA, pub_key, BIGNUM),
	ASN1_SIMPLE(KCDSA, priv_key, BIGNUM)
} ASN1_SEQUENCE_END_cb(KCDSA, KCDSAPrivateKey)

IMPLEMENT_ASN1_ENCODE_FUNCTIONS_const_fname(KCDSA, KCDSAPrivateKey, KCDSAPrivateKey)

ASN1_SEQUENCE_cb(KCDSAparams, kcdsa_cb) = {
	ASN1_SIMPLE(KCDSA, p, BIGNUM),
	ASN1_SIMPLE(KCDSA, q, BIGNUM),
	ASN1_SIMPLE(KCDSA, g, BIGNUM),
} ASN1_SEQUENCE_END_cb(KCDSA, KCDSAparams)

IMPLEMENT_ASN1_ENCODE_FUNCTIONS_const_fname(KCDSA, KCDSAparams, KCDSAparams)

/* DSA public key is a bit trickier... its effectively a CHOICE type
 * decided by a field called write_params which can either write out
 * just the public key as an INTEGER or the parameters and public key
 * in a SEQUENCE
 */
/* The same comment applies to KCDSA also. */

ASN1_SEQUENCE(kcdsa_pub_internal) = {
	ASN1_SIMPLE(KCDSA, pub_key, BIGNUM),
	ASN1_SIMPLE(KCDSA, p, BIGNUM),
	ASN1_SIMPLE(KCDSA, q, BIGNUM),
	ASN1_SIMPLE(KCDSA, g, BIGNUM)
} ASN1_SEQUENCE_END_name(KCDSA, kcdsa_pub_internal)

ASN1_CHOICE_cb(KCDSAPublicKey, kcdsa_cb) = {
	ASN1_SIMPLE(KCDSA, pub_key, BIGNUM),
	ASN1_EX_COMBINE(0, 0, kcdsa_pub_internal)
} ASN1_CHOICE_END_cb(KCDSA, KCDSAPublicKey, write_params)

IMPLEMENT_ASN1_ENCODE_FUNCTIONS_const_fname(KCDSA, KCDSAPublicKey, KCDSAPublicKey)

int KCDSA_sign(int type, const unsigned char *mesg, int mlen, unsigned char *sig,
	     unsigned int *siglen, KCDSA *kcdsa)
	{
	KCDSA_SIG *s;
	s=KCDSA_do_sign(mesg,mlen,kcdsa);
	if (s == NULL)
		{
		*siglen=0;
		return(0);
		}
	*siglen=i2d_KCDSA_SIG(s,&sig);
	KCDSA_SIG_free(s);
	return(1);
	}

int KCDSA_size(const KCDSA *r)
	{
	int ret,i;
	ASN1_INTEGER bs;
	unsigned char buf[4];	/* 4 bytes looks really small.
				   However, i2d_ASN1_INTEGER() will not look
				   beyond the first byte, as long as the second
				   parameter is NULL. */

	i=BN_num_bits(r->q);
	bs.length=(i+7)/8;
	bs.data=buf;
	bs.type=V_ASN1_INTEGER;
	/* If the top bit is set the asn1 encoding is 1 larger. */
	buf[0]=0xff;	

	i=i2d_ASN1_INTEGER(&bs,NULL);
	i+=i; /* r and s */
	ret=ASN1_object_size(1,i,V_ASN1_SEQUENCE);
	return(ret);
	}

/* data has already been hashed (probably with SHA or SHA-1). */
/* returns
 *      1: correct signature
 *      0: incorrect signature
 *     -1: error
 */
int KCDSA_verify(int type, const unsigned char *mesg, int mlen,
	     const unsigned char *sigbuf, int siglen, KCDSA *kcdsa)
	{
	KCDSA_SIG *s;
	int ret=-1;

	s = KCDSA_SIG_new();
	if (s == NULL) return(ret);
	if (d2i_KCDSA_SIG(&s,&sigbuf,siglen) == NULL) goto err;
	ret=KCDSA_do_verify(mesg,mlen,s,kcdsa);
err:
	KCDSA_SIG_free(s);
	return(ret);
	}

