/* kcdsa/kcdsa_iso_paramgen.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include "kcdsa_lcl.h"

#include <string.h>
#include <openssl/evp.h>
#include <openssl/bn.h>
#include <openssl/rand.h>

/* KCDSA의 parameter generation은 p, q, g를 생성하는 것인데 ISO/IEC 14888-3 2nd
 * ed.에서는 KCDSA를 위한 별도의 parameter generation에 대해 언급하고 있지
 * 않으며 FIPS와 호환되는 parameter generation이 기술되어 있다. 따라서 TTA
 * 표준에 얽메일 필요가 없기 때문에 여기서는 14888-3의 parameter generation을
 * 따르기로 한다. 원래 openssl-0.9.8k에 있던 parameter generation은 약간 맞지
 * 않는 면이 있기 때문에 수정하였다. */
/* parameter generation을 위해서도 hash function이 사용되는데 14888-3에서는
 * KCDSA를 위한 별도의 hash function을 제약하고 있지 않기 때문에 evp_md를
 * 입력으로 받도록 한다. */
int kcdsa_iso_paramgen(KCDSA *ret, int pbits, int qbits,
		unsigned char *seed_in, int seed_len, const EVP_MD *evp_md,
		int *counter_ret, unsigned long *h_ret, BN_GENCB *cb)
{
	int ok=0;
	/* TODO: hash function 관련한 것은 나중에 한꺼번에 고치기로 함. */
	unsigned char seed[KCDSA_MAX_Q_LENGTH];
	unsigned char md[KCDSA_MAX_Q_LENGTH];
	unsigned char buf[KCDSA_MAX_Q_LENGTH];
	BIGNUM *W, *tmp, *r0, *c, *X;
	BIGNUM *g=NULL,*q=NULL,*p=NULL;
	BN_MONT_CTX *mont=NULL;
	int k,n=0,i,b,m=0;
	int counter=0;
	int r=0;
	BN_CTX *ctx=NULL;
	unsigned int h=2;
	size_t qlen, md_len;
	int bad_seed = 0;

	if (evp_md == NULL) evp_md = EVP_sha1();
	ret->evp_md = evp_md;

	/* INFO: kcdsa에서는 1024-bit 부터 256-bit 단위로 증가함. max: 2048-bit */
	if (pbits < 1024) pbits=1024;
	pbits=(pbits+255)/256*256;
	/* q 길이 조정 */
	qlen = (qbits+7)/8;
	/* seed를 새로 생성할 경우 size는 qlen이 될 것이므로 이 크기가
	 * KCDSA_MAX_Q_LENGTH보다 커서는 안된다. */
	if (qlen > KCDSA_MAX_Q_LENGTH)
		qlen = KCDSA_MAX_Q_LENGTH;
	md_len = EVP_MD_size(evp_md);
	/* qlen이 md_len보다 크면 hash function의 안전도를 충분히 활용하지 못하게
	 * 된다. 이때는 abort할 수도 있으나 qlen을 md_len으로 줄이고 나머지 과정을
	 * 진행한다. */
	if (qlen > md_len)
		qlen = md_len;

	/* NB: seed_len == 0 is special case: copy generated seed to
 	 * seed_in if it is not NULL.
 	 */
	if (seed_len && (seed_len < qlen))
		seed_in = NULL; /* seed buffer too small -- ignore */
	if (seed_len > KCDSA_MAX_Q_LENGTH)
		seed_len = KCDSA_MAX_Q_LENGTH;
	if (seed_len && (seed_in != NULL))
		{
		memcpy(seed,seed_in,seed_len);
		/* set seed_in to NULL to avoid it being copied back */
		seed_in = NULL;
		}

	if ((ctx=BN_CTX_new()) == NULL) goto err;

	if ((mont=BN_MONT_CTX_new()) == NULL) goto err;

	BN_CTX_start(ctx);
	r0 = BN_CTX_get(ctx);
	g = BN_CTX_get(ctx);
	W = BN_CTX_get(ctx);
	q = BN_CTX_get(ctx);
	X = BN_CTX_get(ctx);
	c = BN_CTX_get(ctx);
	p = BN_CTX_get(ctx);
	tmp = BN_CTX_get(ctx);

	if (!BN_lshift(tmp,BN_value_one(),pbits-1))
		goto err;

	for (;;)
		{
		for (;;) /* find q */
			{
			int seed_is_random;

			/* step 1 */

			/* INFO: seed_len이 0일 경우 seed 입력이 없는 경우이고, 이 때는
			 * seed를 random하게 선택하는 것으로 시작한다. */
			if ((!seed_len) || (bad_seed))
				{
				if (!BN_GENCB_call(cb, 0, m++)) goto err;
				RAND_pseudo_bytes(seed,qlen);
				seed_is_random = 1;
				seed_len = qlen;
				}
			else
				{
				seed_is_random = 0;
				}

			/* step 2 */
			EVP_Digest(seed,seed_len,md,NULL,evp_md, NULL);
			if (!BN_bin2bn(md,qlen,q)) goto err;
			if (BN_num_bits(q) > qbits) 
				if (!BN_mask_bits(q, qbits)) goto err;
			
			/* step 3 */
			BN_set_bit(q, qbits-1);
			BN_set_bit(q, 0);

			/* step 4 */
			r = BN_is_prime_fasttest_ex(q, kcdsa_num_prime_checks, ctx,
					seed_is_random, cb);
			if (r > 0)
				break;
			if (r != 0)
				goto err;

			/* step 5 */
			bad_seed=1; /* The seed was bad. */
			}

		if(!BN_GENCB_call(cb, 2, 0)) goto err;
		if(!BN_GENCB_call(cb, 3, 0)) goto err;

		/* step 6 */
		counter=0;
		/* "offset = 1" */

		n=(pbits-1)/qbits;
		b=(pbits-1)-n*qbits;

		memcpy(buf,seed,seed_len);

		for (;;) /* find p */
			{
			if ((counter) && !BN_GENCB_call(cb, 0, counter))
				goto err;

			/* step 7 */
			BN_zero(W);
			/* now 'buf' contains "SEED + offset - 1" */
			for (k=0; k<=n; k++)
				{
				/* NB: 여기에서 사용하는 방식은 FIPS186과 14888-3에서 사용하고
				 * 있는 방식인데, TTAS.KO-12.0001/R1 (KCDSA)에서 사용하고 있는
				 * 것과 다르다. 또, 14888-3에서 제시하고 있는 것도 DSA domain
				 * parameter를 생성하기 위한 것이고 KCDSA나 다른 알고리즘들의
				 * 경우에 어떻게 할 것인지는 기술하지 않고 있다. */
				/* obtain "SEED + offset + k" by incrementing: */
				for (i=seed_len-1; i >= 0; i--)
					{
					buf[i]++;
					if (buf[i] != 0) break;
					}

				EVP_Digest(buf,seed_len,md,NULL,evp_md, NULL);

				/* step 8 */
				if (!BN_bin2bn(md,md_len,r0))
					goto err;
				if (!BN_lshift(r0,r0,md_len*8*k)) goto err;
				if (!BN_add(W,W,r0)) goto err;
				}

			/* more of step 8 */
			if (BN_num_bits(W) > pbits-1) 
				if (!BN_mask_bits(W,pbits-1)) goto err;
			if (!BN_copy(X,W)) goto err;
			if (!BN_add(X,X,tmp)) goto err;

			/* step 9 */
			if (!BN_lshift1(r0,q)) goto err;
			if (!BN_mod(c,X,r0,ctx)) goto err;
			if (!BN_sub(r0,c,BN_value_one())) goto err;
			if (!BN_sub(p,X,r0)) goto err;

			/* step 10 */
			if (BN_cmp(p,tmp) >= 0)
				{
				/* step 11 */
				r = BN_is_prime_fasttest_ex(p, kcdsa_num_prime_checks,
						ctx, 1, cb);
				if (r > 0)
						goto end; /* found it */
				if (r != 0)
					goto err;
				}

			/* step 13 */
			counter++;
			/* "offset = offset + n + 1" */

			/* step 14 */
			if (counter >= 4096) break;
			}
		}
end:
	if(!BN_GENCB_call(cb, 2, 1))
		goto err;

	/* INFO: 아래 방법은 14888-3 Ann. D.2에 있는 두 가지의 방법 중 Unverifiable
	 * generation이다. F = 2로부터 시작하여 F^r0 mod p를 구하고 g를 못 찾으면
	 * F를 increment하고 다시 시도한다. 14888-3에서는 F에 특별한 제약은 없고
	 * 이전에 시도하지 않은 값을 사용하도록 하고 있다. */
	/* We now need to generate g */
	/* Set r0=(p-1)/q */
	if (!BN_sub(tmp,p,BN_value_one())) goto err;
	if (!BN_div(r0,NULL,tmp,q,ctx)) goto err;

	if (!BN_set_word(tmp,h)) goto err;
	if (!BN_MONT_CTX_set(mont,p,ctx)) goto err;

	for (;;)
		{
		/* g=tmp^r0%p */
		if (!BN_mod_exp_mont(g,tmp,r0,p,ctx,mont)) goto err;
		if (!BN_is_one(g)) break;
		if (!BN_add(tmp,tmp,BN_value_one())) goto err;
		h++;
		}

	if(!BN_GENCB_call(cb, 3, 1))
		goto err;

	ok=1;
err:
	if (ok)
		{
		if(ret->p) BN_free(ret->p);
		if(ret->q) BN_free(ret->q);
		if(ret->g) BN_free(ret->g);
		ret->p=BN_dup(p);
		ret->q=BN_dup(q);
		ret->g=BN_dup(g);
		if (ret->p == NULL || ret->q == NULL || ret->g == NULL)
			{
			ok=0;
			goto err;
			}
		if (seed_in != NULL) memcpy(seed_in,seed,20);
		if (counter_ret != NULL) *counter_ret=counter;
		if (h_ret != NULL) *h_ret=h;
		}
	if(ctx)
		{
		BN_CTX_end(ctx);
		BN_CTX_free(ctx);
		}
	if (mont != NULL) BN_MONT_CTX_free(mont);
	return ok;
}

