/* kcdsa/kcdsa_iso_sign.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 * minor modification from DSA source code
 */

#include "kcdsa_lcl.h"

#include <string.h>
#include <openssl/evp.h>

KCDSA_SIG *kcdsa_do_sign(const unsigned char *mesg, int mlen, KCDSA *kcdsa)
{
	BIGNUM *k=NULL,*s=NULL;
	unsigned char *r = NULL;
	size_t rlen = 0;
	BIGNUM YY;
	BIGNUM V;
	BN_CTX *ctx=NULL;
	int i,reason=ERR_R_BN_LIB;
	KCDSA_SIG *ret=NULL;
	const EVP_MD *evp_md = kcdsa->evp_md;
	EVP_MD_CTX *md_ctx = NULL;
	unsigned char buf[1024]; /* TODO: 좀 경제적으로 */
	size_t mdlen;

	BN_init(&YY);
	BN_init(&V);

	if (!kcdsa->p || !kcdsa->q || !kcdsa->g)
		{
		reason=KCDSA_R_MISSING_PARAMETERS;
		goto err;
		}

	s=BN_new();
	if (s == NULL) goto err;

	ctx=BN_CTX_new();
	if (ctx == NULL) goto err;

	if ((kcdsa->k == NULL) || (kcdsa->r == NULL))
		{
		if (!KCDSA_sign_setup(kcdsa,ctx,&k,&r,&rlen)) goto err;
		}
	else
		{
		k=kcdsa->k;
		kcdsa->k=NULL;
		r=kcdsa->r;
		kcdsa->r=NULL;
		rlen = kcdsa->rlen;
		kcdsa->rlen = 0;
		}

	/* 14888-3 step 6.2.3.5 */ /* TTA 단계 4 */
	/* get Y' = I2BS(Y mod 2^l) */
	if (BN_copy(&YY, kcdsa->pub_key) == NULL) goto err;
	if (!BN_mask_bits(&YY, EVP_MD_block_size(evp_md)*8)) goto err;
	i = BN_bn2bin(&YY, buf); /* TODO: fixed length여야 한다. 길이는 l */
	/* Y'의 길이가 0이 되는 경우는 확률상 0에 가깝다. */
	if (i == 0) goto err;

	/* 14888-3 step 6.2.3.5 계속 */ /* TTA 단계 4 계속 */
	/* md init */
	md_ctx = EVP_MD_CTX_create();
	if (md_ctx == NULL) goto err;
	if (!EVP_DigestInit_ex(md_ctx, evp_md, NULL)) goto err;
	/* update with Y' */
	if (!EVP_DigestUpdate(md_ctx, buf, i)) goto err;
	/* update with mesg */
	if (!EVP_DigestUpdate(md_ctx, mesg, mlen)) goto err;
	/* md final */
	if (!EVP_DigestFinal_ex(md_ctx, buf, &mdlen)) goto err;
	/* now we got md */

	/* 14888-3 step 6.2.3.5 계속 */ /* TTA 단계 5 */
	/* V = BS2I(r xor md) mod q */
	/* TODO: assertion! */
#if 0
	rlen = strlen(r);
	if (rlen != mdlen) goto err;
#endif
	for (i = mdlen-1; i >= 0; i--)
	{
		buf[i] ^= r[i];
	}
	if (!BN_bin2bn(buf, mdlen, &YY)) goto err;
	if (!BN_mod(&V, &YY, kcdsa->q, ctx)) goto err;

	/* 14888-3 step 6.2.3.6 */ /* TTA 단계 6 */
	/* s = x(k - V) mod q */
	while (BN_cmp(k,&V) < 0)
		BN_add(k,k,kcdsa->q);
	BN_sub(s,k,&V);
	if (!BN_mod_mul(s, s, kcdsa->priv_key, kcdsa->q, ctx)) goto err;

	/* 14888-3 step 6.2.3.7
	 * 실제로 appendix를 구성하는 것은 다른 곳에서 이루어져야 한다. 여기에서는
	 * 일단 KCDSA_SIG structure에 값들을 저장하는 것으로 마무리 한다. */
	ret=KCDSA_SIG_new();
	if (ret == NULL) goto err;
	ret->r = r;
	ret->rlen = rlen;
	ret->s = s;
	
err:
	if (!ret)
		{
		//KCDSAerr(KCDSA_F_KCDSA_DO_SIGN,reason);
		if (r != NULL) CRYPTO_free(r);
		if (s != NULL) BN_clear_free(s);
		fprintf(stderr, "signing failed\n");
		}
	if (ctx != NULL) BN_CTX_free(ctx);
	if (md_ctx != NULL) EVP_MD_CTX_destroy(md_ctx);
	BN_clear_free(&YY);
	BN_clear_free(&V);
	if (k != NULL) BN_clear_free(k);
	return(ret);
}

int kcdsa_sign_setup(KCDSA *kcdsa, BN_CTX *ctx_in, BIGNUM **kp, unsigned char **rp, size_t *rlenp)
{
	BN_CTX *ctx = NULL;
	BIGNUM *k=NULL,kq,*K,*R=NULL;
	int i,ret=0;
	unsigned char buf[(KCDSA_MAX_MODULUS_BITS+7)/8];
	unsigned char *r;
	size_t rlen;
	const EVP_MD *evp_md = kcdsa->evp_md;

	if (!kcdsa->p || !kcdsa->q || !kcdsa->g)
		{
		//KCDSAerr(KCDSA_F_KCDSA_SIGN_SETUP,KCDSA_R_MISSING_PARAMETERS);
		return 0;
		}

	BN_init(&kq);

	if (ctx_in == NULL)
		{
		if ((ctx=BN_CTX_new()) == NULL) goto err;
		}
	else
		ctx=ctx_in;

	if ((R=BN_new()) == NULL) goto err;
	if ((k=BN_new()) == NULL) goto err;

	/* 14888-3 step 6.2.3.2 */ /* TTA 단계 1 */
	/* Get random k */
	do
		if (!BN_rand_range(k, kcdsa->q)) goto err;
	while (BN_is_zero(k));
	if ((kcdsa->flags & KCDSA_FLAG_NO_EXP_CONSTTIME) == 0)
		{
		BN_set_flags(k, BN_FLG_CONSTTIME);
		}
	/* TTA 단계 2 */
	/* Compute r = g^k mod p */
	if ((kcdsa->flags & KCDSA_FLAG_NO_EXP_CONSTTIME) == 0)
		{
		if (!BN_copy(&kq, k)) goto err;

		/* 다음의 방법은 timing은 감추지만 power analysis에는 여전히 취약하다.
		 * 그래도 없는 것 보다는 나으니 그냥 두기로 한다. */
		/* We do not want timing information to leak the length of k,
		 * so we compute g^k using an equivalent exponent of fixed length.
		 *
		 * (This is a kludge that we need because the BN_mod_exp_mont()
		 * does not let us specify the desired timing behaviour.) */

		if (!BN_add(&kq, &kq, kcdsa->q)) goto err;
		if (BN_num_bits(&kq) <= BN_num_bits(kcdsa->q))
			{
			if (!BN_add(&kq, &kq, kcdsa->q)) goto err;
			}

		K = &kq;
		}
	else
		{
		K = k;
		}
	if (!BN_mod_exp_mont(R, kcdsa->g, K, kcdsa->p, ctx, kcdsa->method_mont_p))
		goto err;

	/* 14888-3 step 6.2.3.4 */ /* TTA 단계 3 */
	/* compute h(I2BS(R)) */
	i = BN_bn2bin(R, buf);
	if (i == 0) goto err;
	if (!EVP_Digest(buf, i, buf, &rlen, evp_md, NULL)) goto err;
	/* r로 복사 */
	r = CRYPTO_malloc(rlen, __FILE__, __LINE__);
	if (r == NULL) goto err;
	memcpy(r, buf, rlen);

	if (*kp != NULL) BN_clear_free(*kp);
	*kp=k;
	k=NULL;
	if (*rp != NULL) {
		OPENSSL_free(*rp);
		}
	*rp=r;
	r = NULL;
	*rlenp = rlen;
	ret=1;

err:
	if (!ret)
		{
		//KCDSAerr(KCDSA_F_KCDSA_SIGN_SETUP,ERR_R_BN_LIB);
		}
	if (ctx_in == NULL) BN_CTX_free(ctx);
	if (k != NULL) BN_clear_free(k);
	if (R != NULL) BN_clear_free(R);
	if (r != NULL) OPENSSL_free(r);
	BN_clear_free(&kq);
	return(ret);
}

