/* kcdsa/kcdsa_iso_vrf.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include "kcdsa_lcl.h"

#include <string.h>
#include <openssl/evp.h>

/* return 0 on invalid signature
 * 		  1 on valid signature
 * 		  -1 on error
 */
int kcdsa_do_verify(const unsigned char *mesg, int mlen,
		KCDSA_SIG *sig, KCDSA *kcdsa)
{
	BN_CTX *ctx;
	BIGNUM YY, V, R;
	BIGNUM t1;
	// TODO: check the following
	//BN_MONT_CTX *mont=NULL;
	int i, ret = -1;
	const EVP_MD *evp_md = kcdsa->evp_md;
	EVP_MD_CTX *md_ctx = NULL;
	unsigned char buf[1024]; /* TODO: 좀 경제적으로 */
	size_t rlen, mdlen;

	if (!kcdsa->p || !kcdsa->q || !kcdsa->g)
		{
		//KCDSAerr(KCDSA_F_KCDSA_DO_VERIFY,KCDSA_R_MISSING_PARAMETERS);
		return -1;
		}

	if (BN_num_bits(kcdsa->q) > KCDSA_MAX_Q_LENGTH*8)
		{
		//KCDSAerr(KCDSA_F_KCDSA_DO_VERIFY,KCDSA_R_BAD_Q_VALUE);
		return -1;
		}

	if (BN_num_bits(kcdsa->p) > KCDSA_MAX_MODULUS_BITS)
		{
		//KCDSAerr(KCDSA_F_KCDSA_DO_VERIFY,KCDSA_R_MODULUS_TOO_LARGE);
		return -1;
		}

	BN_init(&YY);
	BN_init(&R); // TODO: is it necessary?
	BN_init(&V);
	BN_init(&t1);

	if ((ctx=BN_CTX_new()) == NULL) goto err;

	/* 서명의 크기 검증 */
	if (sig->rlen != EVP_MD_size(evp_md))
		{
		ret = 0;
		goto err;
		}
	if (BN_is_zero(sig->s) || BN_is_negative(sig->s) ||
	    BN_ucmp(sig->s, kcdsa->q) >= 0)
		{
		ret = 0;
		goto err;
		}

	/* 아래 과정은 14888-3 step 6.2.4.3이지만 6.2.3.5와 동일하다 */
	/* 14888-3 step 6.2.3.5 */
	/* get Y' = I2BS(Y mod 2^l) */
	if (BN_copy(&YY, kcdsa->pub_key) == NULL) goto err;
	if (!BN_mask_bits(&YY, EVP_MD_block_size(evp_md)*8)) goto err;
	i = BN_bn2bin(&YY, buf);
	/* Y'의 길이가 0이 되는 경우는 확률상 0에 가깝다. */
	if (i == 0) goto err;

	/* 14888-3 step 6.2.3.5 계속 */
	/* md init */
	md_ctx = EVP_MD_CTX_create();
	if (md_ctx == NULL) goto err;
	if (!EVP_DigestInit_ex(md_ctx, evp_md, NULL)) goto err;
	/* update with Y' */
	if (!EVP_DigestUpdate(md_ctx, buf, i)) goto err;
	/* update with mesg */
	if (!EVP_DigestUpdate(md_ctx, mesg, mlen)) goto err;
	/* md final */
	if (!EVP_DigestFinal_ex(md_ctx, buf, &mdlen)) goto err;
	/* now we got md */

	/* 14888-3 step 6.2.3.5 계속 */
	/* V = BS2I(r xor md) mod q */
	/* TODO: assertion! */
	rlen = sig->rlen;
	if (rlen != mdlen) goto err;
	for (i = mdlen-1; i >= 0; i--)
	{
		buf[i] ^= sig->r[i];
	}
	if (!BN_bin2bn(buf, mdlen, &YY)) goto err;
	if (!BN_mod(&V, &YY, kcdsa->q, ctx)) goto err;

	/* 14888-3 step 6.2.4.4 */
	/* Y^s * g^v mod p */
	if (!BN_mod_exp2_mont(&t1,
				kcdsa->pub_key, sig->s,
				kcdsa->g, &V,
				kcdsa->p, ctx, kcdsa->method_mont_p))
		goto err;

	/* 아래 과정은 14888-3 step 6.2.4.5이지만 6.2.3.4와 동일하다 */
	/* 14888-3 step 6.2.3.4 */
	/* compute h(I2BS(R)) */
	i = BN_bn2bin(&t1, buf);
	if (i == 0) goto err;
	if (!EVP_Digest(buf, i, buf, &rlen, evp_md, NULL)) goto err;

	/* 14888-3 step 6.2.4.6 */
	if (memcmp(buf, sig->r, rlen)) 
	{
		ret = 0;
		goto err;
	}

	ret = 1;

err:
	//if (ret == -1) KCDSAerr(KCDSA_F_KCDSA_DO_VERIFY,ERR_R_BN_LIB);
	if (ctx != NULL) BN_CTX_free(ctx);
	if (md_ctx != NULL) EVP_MD_CTX_destroy(md_ctx);
	BN_clear_free(&YY);
	BN_clear_free(&R);
	BN_clear_free(&V);
	BN_clear_free(&t1);
	return(ret);
}

