/* kcdsa/kcdsa_lib.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 * openssl-0.9.8k에 있는 것을 가져와서 약간 수정함
 */

#include <stdio.h>
#include <openssl/bn.h>
#include "kcdsa_lcl.h"
#include <openssl/asn1.h>
#ifndef OPENSSL_NO_DH
#include <openssl/dh.h>
#endif

/* originally in crypto/asn1/t_pkey.c */
static int print(BIO *fp,const char *str, const BIGNUM *num,
		unsigned char *buf,int off);

const char KCDSA_version[]="KCDSA";

static const KCDSA_METHOD *default_KCDSA_method = NULL;

void KCDSA_set_default_method(const KCDSA_METHOD *meth)
	{
	default_KCDSA_method = meth;
	}

const KCDSA_METHOD *KCDSA_get_default_method(void)
	{
	if(!default_KCDSA_method)
		default_KCDSA_method = KCDSA_iso_method();
	return default_KCDSA_method;
	}

KCDSA *KCDSA_new(void)
	{
	return KCDSA_new_method(NULL);
	}

int KCDSA_set_method(KCDSA *kcdsa, const KCDSA_METHOD *meth)
	{
	/* NB: The caller is specifically setting a method, so it's not up to us
	 * to deal with which ENGINE it comes from. */
        const KCDSA_METHOD *mtmp;
        mtmp = kcdsa->meth;
        if (mtmp->finish) mtmp->finish(kcdsa);
        kcdsa->meth = meth;
        if (meth->init) meth->init(kcdsa);
        return 1;
	}

KCDSA *KCDSA_new_method(ENGINE *engine)
	{
	KCDSA *ret;

	ret=(KCDSA *)OPENSSL_malloc(sizeof(KCDSA));
	if (ret == NULL)
		{
		//KCDSAerr(KCDSA_F_KCDSA_NEW_METHOD,ERR_R_MALLOC_FAILURE);
		return(NULL);
		}
	ret->meth = KCDSA_get_default_method();

	ret->pad=0;
	ret->version=0;
	ret->write_params=1;
	ret->p=NULL;
	ret->q=NULL;
	ret->g=NULL;
	ret->evp_md = NULL;

	ret->pub_key=NULL;
	ret->priv_key=NULL;

	ret->k=NULL;
	ret->r=NULL;
	ret->rlen = 0;
	ret->method_mont_p=NULL;

	//ret->references=1;
	ret->flags=ret->meth->flags;
	//CRYPTO_new_ex_data(CRYPTO_EX_INDEX_KCDSA, ret, &ret->ex_data);
	if ((ret->meth->init != NULL) && !ret->meth->init(ret))
		{
		/* TODO: CRYPTO_EX_INDEX_KCDSA */
		//CRYPTO_free_ex_data(CRYPTO_EX_INDEX_KCDSA, ret, &ret->ex_data);
		OPENSSL_free(ret);
		ret=NULL;
		}
	
	return(ret);
	}

void KCDSA_free(KCDSA *r)
	{
	if (r == NULL) return;

	if(r->meth->finish)
		r->meth->finish(r);

	if (r->p != NULL) BN_clear_free(r->p);
	if (r->q != NULL) BN_clear_free(r->q);
	if (r->g != NULL) BN_clear_free(r->g);
	if (r->evp_md != NULL) r->evp_md = NULL; /* evp_md contains a pointer to a
												static structure. don't free
												it. just it null. */
	if (r->pub_key != NULL) BN_clear_free(r->pub_key);
	if (r->priv_key != NULL) BN_clear_free(r->priv_key);
	if (r->k != NULL) BN_clear_free(r->k);
	if (r->r != NULL) {
		OPENSSL_cleanse(r->r, r->rlen);
		OPENSSL_free(r->r);
		}
	OPENSSL_free(r);
	}

int KCDSA_do_verify(const unsigned char *mesg, int mlen, KCDSA_SIG *sig,
		KCDSA *kcdsa)
	{
	return kcdsa->meth->kcdsa_do_verify(mesg, mlen, sig, kcdsa);
	}

int KCDSA_generate_parameters_ex(KCDSA *kcdsa, int pbits, int qbits,
		unsigned char *seed_in, int seed_len, const EVP_MD *evp_md,
		int *counter_ret, unsigned long *h_ret, BN_GENCB *cb)
	{
	return kcdsa->meth->kcdsa_paramgen(kcdsa, pbits, qbits, seed_in, seed_len,
				evp_md, counter_ret, h_ret, cb);
	}

int KCDSA_generate_key(KCDSA *kcdsa)
	{
	return kcdsa->meth->kcdsa_keygen(kcdsa);
	}

KCDSA_SIG * KCDSA_do_sign(const unsigned char *mesg, int mlen, KCDSA *kcdsa)
	{
	return kcdsa->meth->kcdsa_do_sign(mesg, mlen, kcdsa);
	}

int KCDSA_sign_setup(KCDSA *kcdsa, BN_CTX *ctx_in,
		BIGNUM **kp, unsigned char **rp, size_t *rlenp)
	{
	return kcdsa->meth->kcdsa_sign_setup(kcdsa, ctx_in, kp, rp, rlenp);
	}

KCDSA_SIG *KCDSA_SIG_new(void)
	{
	KCDSA_SIG *sig;
	sig = OPENSSL_malloc(sizeof(KCDSA_SIG));
	if (!sig)
		return NULL;
	sig->r = NULL;
	sig->rlen = 0;
	sig->s = NULL;
	return sig;
	}

void KCDSA_SIG_free(KCDSA_SIG *sig)
	{
	if (sig)
		{
		if (sig->r)
			OPENSSL_free(sig->r);
		if (sig->rlen)
			sig->rlen = 0;
		if (sig->s)
			BN_free(sig->s);
		OPENSSL_free(sig);
		}
	}

/* following functions originally in crypto/asn1/t_pkey.c */
#ifndef OPENSSL_NO_BIO
int	KCDSAparams_print(BIO *bp, const KCDSA *x)
{
	return 0;
}

int	KCDSA_print(BIO *bp, const KCDSA *x, int off)
{
	unsigned char *m=NULL;
	int ret=0;
	size_t buf_len=0,i;

	if (x->p)
		buf_len = (size_t)BN_num_bytes(x->p);
	else
		{
		//DSAerr(DSA_F_DSA_PRINT,DSA_R_MISSING_PARAMETERS);
		goto err;
		}
	if (x->q)
		if (buf_len < (i = (size_t)BN_num_bytes(x->q)))
			buf_len = i;
	if (x->g)
		if (buf_len < (i = (size_t)BN_num_bytes(x->g)))
			buf_len = i;
	if (x->priv_key)
		if (buf_len < (i = (size_t)BN_num_bytes(x->priv_key)))
			buf_len = i;
	if (x->pub_key)
		if (buf_len < (i = (size_t)BN_num_bytes(x->pub_key)))
			buf_len = i;

	m=(unsigned char *)OPENSSL_malloc(buf_len+10);
	if (m == NULL)
		{
		//DSAerr(DSA_F_DSA_PRINT,ERR_R_MALLOC_FAILURE);
		goto err;
		}

	if (x->priv_key != NULL)
		{
		if(!BIO_indent(bp,off,128))
		   goto err;
		if (BIO_printf(bp,"Private-Key: (%d bit)\n",BN_num_bits(x->p))
			<= 0) goto err;
		}

	if ((x->priv_key != NULL) && !print(bp,"priv:",x->priv_key,m,off))
		goto err;
	if ((x->pub_key  != NULL) && !print(bp,"pub: ",x->pub_key,m,off))
		goto err;
	if ((x->p != NULL) && !print(bp,"P:   ",x->p,m,off)) goto err;
	if ((x->q != NULL) && !print(bp,"Q:   ",x->q,m,off)) goto err;
	if ((x->g != NULL) && !print(bp,"G:   ",x->g,m,off)) goto err;
	ret=1;
err:
	if (m != NULL) OPENSSL_free(m);
	return(ret);
}
#endif

#ifndef OPENSSL_NO_FP_API
int	KCDSAparams_print_fp(FILE *fp, const KCDSA *x)
{
	return 0;
}

int	KCDSA_print_fp(FILE *fp, const KCDSA *x, int off)
{
	BIO *b;
	int ret;

	if ((b=BIO_new(BIO_s_file())) == NULL)
		{
		//KCDSAerr(DSA_F_KCDSA_PRINT_FP,ERR_R_BUF_LIB);
		return(0);
		}
	BIO_set_fp(b,fp,BIO_NOCLOSE);
	ret=KCDSA_print(b,x,off);
	BIO_free(b);
	return(ret);
}
#endif

static int print(BIO *bp, const char *number, const BIGNUM *num, unsigned char *buf,
	     int off)
	{
	int n,i;
	const char *neg;

	if (num == NULL) return(1);
	neg = (BN_is_negative(num))?"-":"";
	if(!BIO_indent(bp,off,128))
		return 0;
	if (BN_is_zero(num))
		{
		if (BIO_printf(bp, "%s 0\n", number) <= 0)
			return 0;
		return 1;
		}

	if (BN_num_bytes(num) <= BN_BYTES)
		{
		if (BIO_printf(bp,"%s %s%lu (%s0x%lx)\n",number,neg,
			(unsigned long)num->d[0],neg,(unsigned long)num->d[0])
			<= 0) return(0);
		}
	else
		{
		buf[0]=0;
		if (BIO_printf(bp,"%s%s",number,
			(neg[0] == '-')?" (Negative)":"") <= 0)
			return(0);
		n=BN_bn2bin(num,&buf[1]);
	
		if (buf[1] & 0x80)
			n++;
		else	buf++;

		for (i=0; i<n; i++)
			{
			if ((i%15) == 0)
				{
				if(BIO_puts(bp,"\n") <= 0
				   || !BIO_indent(bp,off+4,128))
				    return 0;
				}
			if (BIO_printf(bp,"%02x%s",buf[i],((i+1) == n)?"":":")
				<= 0) return(0);
			}
		if (BIO_write(bp,"\n",1) <= 0) return(0);
		}
	return(1);
	}
