/* kcdsa/kcdsa_tta_paramgen.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include "kcdsa_lcl.h"

#include <string.h>
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/bn.h>

/* TTAS.KO-12.0001/R1 에서는 모든 과정에서 사용되는 hash function으로 HAS-160을
 * 사용하도록 하고 있으나, 현재 openssl에는 이 hash function이 포함되어 있지
 * 않다. 따라서 ISO method에서와 마찬가지로 function argument로 넘어 오는
 * evp_md를 사용하도록 한다. */
int kcdsa_tta_paramgen(KCDSA *ret, int pbits, int qbits,
		unsigned char *seed_in, int seed_len, const EVP_MD *evp_md,
		int *counter_ret, unsigned long *h_ret, BN_GENCB *cb)
{
	int ok=0;
	/* TODO: hash function 관련한 것은 나중에 한꺼번에 고치기로 함. */
	unsigned char seed[KCDSA_MAX_Q_LENGTH];
	unsigned char buf[KCDSA_MAX_MODULUS_BITS/8];
	BIGNUM *J, *Q, *P, *G;
	BIGNUM *tmp = NULL;
	BN_MONT_CTX *mont=NULL;
	int m = 0;
	int counter=0;
	int r=0;
	BN_CTX *ctx=NULL;
	size_t qlen, md_len;

	if (evp_md == NULL) evp_md = EVP_sha1();
	ret->evp_md = evp_md;

	/* INFO: kcdsa에서는 1024-bit 부터 256-bit 단위로 증가함. max: 2048-bit */
	if (pbits < 1024) pbits=1024;
	pbits=(pbits+255)/256*256;
	/* q 길이 조정 */
	qlen = (qbits+7)/8;
	/* seed를 새로 생성할 경우 size는 qlen이 될 것이므로 이 크기가
	 * KCDSA_MAX_Q_LENGTH보다 커서는 안된다. */
	if (qlen > KCDSA_MAX_Q_LENGTH)
		qlen = KCDSA_MAX_Q_LENGTH;
	md_len = EVP_MD_size(evp_md);
	/* qlen이 md_len보다 크면 hash function의 안전도를 충분히 활용하지 못하게
	 * 된다. 이때는 abort할 수도 있으나 qlen을 md_len으로 줄이고 나머지 과정을
	 * 진행한다. */
	if (qlen > md_len)
		qlen = md_len;

	/* NB: seed_len == 0 is special case: copy generated seed to
 	 * seed_in if it is not NULL.
 	 */
	if (seed_len && (seed_len < qlen))
		seed_in = NULL; /* seed buffer too small -- ignore */
	if (seed_len > KCDSA_MAX_Q_LENGTH)
		seed_len = KCDSA_MAX_Q_LENGTH;
	if ((seed_in != NULL) && (seed_len <= KCDSA_MAX_Q_LENGTH))
		{
		memcpy(seed,seed_in,seed_len);
		/* set seed_in to NULL to avoid it being copied back */
		seed_in = NULL;
		}

	if ((ctx=BN_CTX_new()) == NULL) goto err;

	if ((mont=BN_MONT_CTX_new()) == NULL) goto err;

	BN_CTX_start(ctx);
	J = BN_CTX_get(ctx);
	Q = BN_CTX_get(ctx);
	P = BN_CTX_get(ctx);
	G = BN_CTX_get(ctx);
	tmp = BN_CTX_get(ctx);

	for (;;)
		{
		for (;;)
			{
			int seed_is_random;
			int gen_seed = 0;

			/* 단계 1 */
			if(!BN_GENCB_call(cb, 0, m++))
				goto err;

			/* INFO: seed_len이 0일 경우 seed 입력이 없는 경우이고, 이 때는
			 * seed를 random하게 선택하는 것으로 시작한다. */
			if ((!seed_len) || (gen_seed))
				{
				RAND_pseudo_bytes(seed,qlen);
				seed_is_random = 1;
				seed_len = qlen;
				}
			else
				{
				seed_is_random = 0;
				}
			gen_seed=1;

			/* 단계 2 */
			kcdsa_tta_prng(seed, seed_len, evp_md, buf, pbits-qbits-4);

			/* 단계 3 */
			if (BN_bin2bn(buf, pbits-qbits-4, J) == NULL) goto err;
			if (!BN_set_bit(J, pbits-qbits-1)) goto err;
			if (!BN_set_bit(J, 0)) goto err;

			/* 단계 4 */
			r = BN_is_prime_fasttest_ex(J, kcdsa_num_prime_checks, ctx,
					seed_is_random, cb);
			if (r > 0) break;;
			if (r < 0) goto err;
			}

		/* 단계 5 */
		counter = 0;

		for (;;) {
			/* 단계 6 */
			counter++;
			/* 단계 7 */
			if (counter > (1<<23) ) break;
			/* 단계 8 */
			memcpy(buf, seed, seed_len);
			/* big-endian system이라고 가정한다. */
			buf[seed_len] = (counter>>24) & 0xff;
			buf[seed_len+1] = (counter>>16) & 0xff;
			buf[seed_len+2] = (counter>>8) & 0xff;
			buf[seed_len+3] = counter & 0xff;
			kcdsa_tta_prng(buf, seed_len+4, evp_md, buf, qbits);

			/* 단계 9 */
			if (BN_bin2bn(buf, qbits, Q) == NULL) goto err;
			if (!BN_set_bit(Q, qbits-1)) goto err;
			if (!BN_set_bit(Q, 0)) goto err;

			/* 단계 10 */
			if (!BN_mul(tmp, J, Q, ctx)) goto err;
			if (!BN_lshift1(P, tmp)) goto err;
			if (!BN_add(P, P, BN_value_one())) goto err;
			if (BN_num_bits(P) > pbits) continue;

			r = BN_is_prime_fasttest_ex(Q, kcdsa_num_prime_checks, ctx, 1, cb);
			if (r == 0) continue;;
			if (r < 0) goto err;

			r = BN_is_prime_fasttest_ex(P, kcdsa_num_prime_checks, ctx, 1, cb);
			if (r == 0) continue;;
			if (r < 0) goto err;
			}

		if (counter > (1<<23) ) continue;
		}

//end:
	if(!BN_GENCB_call(cb, 2, 1))
		goto err;

	for (;;) {
		/* 단계 1 */
		BN_pseudo_rand_range(G, P);

		/* 단계 2 */
		if (!BN_MONT_CTX_set(mont,P,ctx)) goto err;
		if (!BN_lshift1(tmp, J)) goto err;
		if (!BN_mod_exp_mont(G,G,tmp,P,ctx,mont)) goto err;
		if (!BN_is_one(G)) break;
		}

	if(!BN_GENCB_call(cb, 3, 1))
		goto err;

	ok=1;
err:
	if (ok)
		{
		if(ret->p) BN_free(ret->p);
		if(ret->q) BN_free(ret->q);
		if(ret->g) BN_free(ret->g);
		ret->p=BN_dup(P);
		ret->q=BN_dup(Q);
		ret->g=BN_dup(G);
		if (ret->p == NULL || ret->q == NULL || ret->g == NULL)
			{
			ok=0;
			goto err;
			}
		if (seed_in != NULL) memcpy(seed_in,seed,seed_len);
		if (counter_ret != NULL) *counter_ret=counter;
		}
	if(ctx)
		{
		BN_CTX_end(ctx);
		BN_CTX_free(ctx);
		}
	if (mont != NULL) BN_MONT_CTX_free(mont);
	return ok;
}

