/* kcdsa/kcdsa_iso_keygen.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 * minor modification from DSA source code
 */

#include "kcdsa_lcl.h"

int kcdsa_iso_keygen(KCDSA *kcdsa)
{
	int ok=0;
	BN_CTX *ctx=NULL;
	BIGNUM *pub_key=NULL,*priv_key=NULL;

	if ((ctx=BN_CTX_new()) == NULL) goto err;

	if (kcdsa->priv_key == NULL)
		{
		if ((priv_key=BN_new()) == NULL) goto err;
		}
	else
		priv_key=kcdsa->priv_key;

	/* NB: TTAS.KO-12.0001/R1에서는 m 명(m >= 1)의 사용자가 사용하기 위한
	 * 비밀키를 만드는 방법이 나와 있으나 통상적으로 m = 1이 되고 이 경우 TTA
	 * 표준의 방식을 사용하든 random number를 바로 사용하든 외부 사용자의
	 * 입장에서는 통계적으로 동일한 효과를 나타낸다. 그러므로 14888-3에서
	 * 제시하듯이 0, q 사이의 random number를 생성하여 바로 비밀키로 사용하는
	 * (DSA와 같은) 방식을 그대로 사용한다. */
	do
		if (!BN_rand_range(priv_key,kcdsa->q)) goto err;
	while (BN_is_zero(priv_key));

	if (kcdsa->pub_key == NULL)
		{
		if ((pub_key=BN_new()) == NULL) goto err;
		}
	else
		pub_key=kcdsa->pub_key;
	
	{
		/* INFO: 원래 DSA 코드에서는 local_prk를 constant time exponentiation을
		 * 사용하기 위한 임시 variable로 사용하고 있다. KCDSA에서는 key pair를
		 * 만드는 방식이 DSA와 다르기 때문에 local_prexp를 prk^{-1} mod p 로
		 * 사용한다. Y = g^local_prexp 이다. */
		BIGNUM local_prk;
		BIGNUM local_prexp; /* private exponent = X^{-1} mod p */
		BIGNUM *prk;

		BN_init(&local_prexp); // TODO: possible memory leak

		if ((kcdsa->flags & KCDSA_FLAG_NO_EXP_CONSTTIME) == 0)
			{
			BN_init(&local_prk); // TODO: possible memory leak
			prk = &local_prk;
			BN_with_flags(prk, priv_key, BN_FLG_CONSTTIME);
			BN_set_flags(&local_prexp, BN_FLG_CONSTTIME);
			}
		else
			prk = priv_key;

		if (!BN_mod_inverse(&local_prexp, prk, kcdsa->q, ctx)) goto err;
		if (!BN_mod_exp(pub_key,kcdsa->g,&local_prexp,kcdsa->p,ctx)) goto err;
	}

	kcdsa->priv_key=priv_key;
	kcdsa->pub_key=pub_key;
	ok=1;

err:
	if ((pub_key != NULL) && (kcdsa->pub_key == NULL)) BN_free(pub_key);
	if ((priv_key != NULL) && (kcdsa->priv_key == NULL)) BN_free(priv_key);
	if (ctx != NULL) BN_CTX_free(ctx);
	return(ok);
}


