/* kcdsa/kcdsa_iso_keygen.c */

/*
 * KCDSA source code by Yeon-Hyeong Yang <yhyang@oberon.postech.ac.kr>
 */

#include "kcdsa_lcl.h"

int kcdsa_tta_keygen(KCDSA *kcdsa)
{
	int ok=0, qbits, len;
	BN_CTX *ctx=NULL;
	BIGNUM *pub_key=NULL,*priv_key=NULL;
	BIGNUM *xkey=NULL, *xseed=NULL;
	const EVP_MD *evp_md = kcdsa->evp_md;
	unsigned char buf[KCDSA_MAX_Q_LENGTH];

	if ((ctx=BN_CTX_new()) == NULL) goto err;

	if (kcdsa->priv_key == NULL)
		{
		if ((priv_key=BN_new()) == NULL) goto err;
		}
	else
		priv_key=kcdsa->priv_key;

	qbits = BN_num_bits(kcdsa->q);

	BN_CTX_start(ctx);
	xkey = BN_CTX_get(ctx);
	xseed = BN_CTX_get(ctx);

	/* NB: TTAS.KO-12.0001/R1에서는 m 명(m >= 1)의 사용자가 사용하기 위한
	 * 비밀키를 만드는 방법이 나와 있으나 통상적으로 m = 1이 된다. xseed는
	 * optional user-provided random input을 prng에 입력으로 넣어서 그 출력을
	 * 취하는 것으로 돼 있으나 일단 BN_rand()로부터 얻기로 한다. */
	/* 4.1 단계 1 */
	if (!BN_rand(xkey, qbits, -1, 0)) goto err;
	/* 4.1 단계 2 */
	/* (a) */
	if (!BN_rand(xseed, qbits, -1, 0)) goto err;
	/* (b) */
	if (!BN_add(xkey, xkey, xseed)) goto err;
	if (!BN_mask_bits(xkey, qbits)) goto err; /* xval */
	/* (d) */
	len = BN_bn2bin(xkey, buf);
	kcdsa_tta_prng(buf, len, evp_md, buf, qbits);
	if (BN_bin2bn(buf, qbits, xkey) == NULL) goto err;
	if (!BN_mod(priv_key, xkey, kcdsa->q, ctx)) goto err;

	if (kcdsa->pub_key == NULL)
		{
		if ((pub_key=BN_new()) == NULL) goto err;
		}
	else
		pub_key=kcdsa->pub_key;
	
	{
		/* INFO: 원래 DSA 코드에서는 local_prk를 constant time exponentiation을
		 * 사용하기 위한 임시 variable로 사용하고 있다. KCDSA에서는 key pair를
		 * 만드는 방식이 DSA와 다르기 때문에 local_prexp를 prk^{-1} mod p 로
		 * 사용한다. Y = g^local_prexp 이다. */
		BIGNUM local_prk;
		BIGNUM local_prexp; /* private exponent = X^{-1} mod p */
		BIGNUM *prk;

		BN_init(&local_prexp);

		if ((kcdsa->flags & KCDSA_FLAG_NO_EXP_CONSTTIME) == 0)
			{
			BN_init(&local_prk);
			prk = &local_prk;
			BN_with_flags(prk, priv_key, BN_FLG_CONSTTIME);
			BN_set_flags(&local_prexp, BN_FLG_CONSTTIME);
			}
		else
			prk = priv_key;

		if (!BN_mod_inverse(&local_prexp, prk, kcdsa->q, ctx)) goto err;
		if (!BN_mod_exp(pub_key,kcdsa->g,&local_prexp,kcdsa->p,ctx)) goto err;
	}

	kcdsa->priv_key=priv_key;
	kcdsa->pub_key=pub_key;
	ok=1;

err:
	if ((pub_key != NULL) && (kcdsa->pub_key == NULL)) BN_free(pub_key);
	if ((priv_key != NULL) && (kcdsa->priv_key == NULL)) BN_free(priv_key);
	if (xkey != NULL) BN_clear(xkey);
	if (xseed != NULL) BN_clear(xseed);
	BN_CTX_end(ctx);
	if (ctx != NULL) BN_CTX_free(ctx);
	return(ok);
}


